#
# Apparmor manual:
# http://manpages.ubuntu.com/manpages/trusty/man5/apparmor.d.5.html
#

#include <tunables/global>

%HOME%%USERNAME%/bin/bash flags=(attach_disconnected) {

  # Just in case (deny overrides allow)
  deny %HOME%%USERNAME%/usr/** wl,
  deny %HOME%%USERNAME%/bin/** wl,
  deny %HOME%%USERNAME%/proc/** wl,
  deny %HOME%%USERNAME%/lib/** wl,
  deny %HOME%%USERNAME%/lib64/** wl,

  # New in ubuntu 18
  %HOME%%USERNAME%/lib/** mr,
  %HOME%%USERNAME%/lib64/** mr,
  %HOME%%USERNAME%/usr/** mr,
  %HOME%%USERNAME%/bin/** mr,

  signal receive set=hup peer=%HOME%%USERNAME%/bin/bash,
  signal receive set=int peer=%HOME%%USERNAME%/bin/bash,
  signal receive set=hup peer=/usr/bin/nodejs_%USERNAME%,
  
  signal send set=int peer=%HOME%%USERNAME%/bin/bash,
  signal send set=hup peer=%HOME%%USERNAME%/bin/bash,
  signal send set=hup peer=%HOME%%USERNAME%/usr/bin/node,
  signal send set=hup peer=%HOME%%USERNAME%/usr/lib/node_modules/npm/bin/npm-cli.js,

  %HOME%%USERNAME%/ r,
  %HOME%%USERNAME%/** mr,
  owner %HOME%%USERNAME%/** wlk,

  %HOME%%USERNAME%/dev/null rw,

  # Connect to other servers via ssh
  network,

  # Able to use the docker service
  # This is probably not a good idea considering Docker has root access,
  # you can for example mount the host system, then modify it from the container
  /run/docker.sock wr,

  # Many scripts wants to access urandom ... For example when doing a SSL handshake!?
  /dev/urandom r,

  # bins that have their own profile
  %HOME%%USERNAME%/usr/bin/hg Px,
  %HOME%%USERNAME%/usr/bin/git Px,
  %HOME%%USERNAME%/usr/bin/node Px,
  %HOME%%USERNAME%/usr/bin/bash Px,
  %HOME%%USERNAME%/usr/bin/python Px,
  %HOME%%USERNAME%/usr/lib/node_modules/npm/bin/npm-cli.js Px,
  %HOME%%USERNAME%/usr/lib/node_modules/npm/bin/npx-cli.js Px,

  # bins without profile (who needs network)
  %HOME%%USERNAME%/usr/local/bin/docker-compose ix,
  %HOME%%USERNAME%/usr/bin/npm ix,
  %HOME%%USERNAME%/usr/bin/env ix,
  %HOME%%USERNAME%/usr/bin/ssh ix,
  %HOME%%USERNAME%/usr/bin/ssh-keygen ix,
  %HOME%%USERNAME%/usr/bin/unzip ix,
  %HOME%%USERNAME%/usr/bin/unrar ix,
  %HOME%%USERNAME%/usr/bin/wget ix,
  %HOME%%USERNAME%/usr/bin/curl ix,
  %HOME%%USERNAME%/usr/bin/host ix,

  %HOME%%USERNAME%/bin/ls ix,
  %HOME%%USERNAME%/bin/tar ix,
  %HOME%%USERNAME%/bin/gunzip ix,
  %HOME%%USERNAME%/bin/sh ix,
  %HOME%%USERNAME%/bin/gzip ix,

  # other bins
  %HOME%%USERNAME%/** Cx -> scripts,

  # this profile should be more restrictive, otherwise put the bin as ix above
  profile scripts {
    deny %HOME%%USERNAME%/usr/** wl,
    deny %HOME%%USERNAME%/bin/** wl,
    deny %HOME%%USERNAME%/proc/** wl,
    deny %HOME%%USERNAME%/lib/** wl,
    deny %HOME%%USERNAME%/lib64/** wl,

    deny %HOME%%USERNAME%/.ssh/** rwl,
    deny %HOME%%USERNAME%/.gnupg/** rwl,

    # Restrict network access! Only allow unix sockets
    network unix,

    %HOME%%USERNAME%/ r,
    %HOME%%USERNAME%/** mr,
    owner %HOME%%USERNAME%/** rwlk,

    # /dev/null is often used
    %HOME%%USERNAME%/dev/null rw,
    /dev/urandom r,

    # Connecting to mySQL
    /run/mysqld/mysqld.sock rw,

    # bins that have their own profile
    %HOME%%USERNAME%/usr/bin/hg Px,
    %HOME%%USERNAME%/usr/bin/git Px,
    %HOME%%USERNAME%/usr/bin/node Px,
    %HOME%%USERNAME%/usr/bin/python Px,
    %HOME%%USERNAME%/usr/lib/node_modules/npm/bin/npm-cli.js Px,
    %HOME%%USERNAME%/usr/lib/node_modules/npm/bin/npx-cli.js Px,

    %HOME%%USERNAME%/bin/bash Px,
    
    %HOME%%USERNAME%/usr/bin/wget px -> networkTool,
    %HOME%%USERNAME%/usr/bin/curl px -> networkTool,
    

    # Scripts of scripts
    %HOME%%USERNAME%/** ix,
   
   
    signal send set=kill peer=%HOME%%USERNAME%/bin/bash//scripts,
    signal send set=kill peer=%HOME%%USERNAME%/bin/bash//networkTool,
    signal send set=abrt peer=%HOME%%USERNAME%/bin/bash//scripts,
    
    signal receive set=kill peer=%HOME%%USERNAME%/bin/bash//scripts,
    signal receive set=abrt peer=%HOME%%USERNAME%/bin/bash//scripts,
 
 
  }

  # Restrictive, but allow network
  profile networkTool {
    deny %HOME%%USERNAME%/usr/** wl,
    deny %HOME%%USERNAME%/bin/** wl,
    deny %HOME%%USERNAME%/proc/** wl,
    deny %HOME%%USERNAME%/lib/** wl,
    deny %HOME%%USERNAME%/lib64/** wl,

    %HOME%%USERNAME%/ r,
    %HOME%%USERNAME%/** mr,
    
    owner %HOME%%USERNAME%/** wl,
    
    %HOME%%USERNAME%/lib/** mr,
    %HOME%%USERNAME%/lib64/** mr,
    %HOME%%USERNAME%/usr/** mr,
    %HOME%%USERNAME%/bin/** mr,
    
    
    %HOME%%USERNAME%/dev/null rw,
    /dev/urandom r,
    
    network,
      
    signal receive set=kill peer=%HOME%%USERNAME%/bin/bash//scripts,
  }



}


